package envoy.server.processors;

import static envoy.data.Message.MessageStatus.*;
import static envoy.server.Startup.config;

import java.time.Instant;
import java.util.Collections;
import java.util.logging.Logger;

import javax.persistence.EntityExistsException;

import envoy.data.GroupMessage;
import envoy.event.*;
import envoy.server.data.PersistenceManager;
import envoy.server.net.*;
import envoy.util.EnvoyLog;

/**
 * @author Maximilian K&auml;fer
 * @since Envoy Server Standalone v0.1-beta
 */
public final class GroupMessageProcessor implements ObjectProcessor<GroupMessage> {

	private static final ConnectionManager	connectionManager	= ConnectionManager.getInstance();
	private static final PersistenceManager	persistenceManager	= PersistenceManager.getInstance();
	private static final Logger				logger				= EnvoyLog.getLogger(GroupCreationProcessor.class);

	@Override
	public void process(GroupMessage groupMessage, long socketID, ObjectWriteProxy writeProxy) {
		groupMessage.nextStatus();

		// Update statuses to SENT / RECEIVED depending on online status
		groupMessage.getMemberStatuses().replaceAll((memberID, status) -> connectionManager.isOnline(memberID) ? RECEIVED : SENT);

		// Set status for sender to READ
		groupMessage.getMemberStatuses().replace(groupMessage.getSenderID(), READ);

		// Increment the overall status to RECEIVED if necessary
		if (Collections.min(groupMessage.getMemberStatuses().values()) == RECEIVED) {
			groupMessage.nextStatus();

			// Notify the sender of the status change
			writeProxy.write(socketID, new MessageStatusChange(groupMessage));
		}

		// message attachment will be automatically removed if disabled in config
		final var groupMessageServer = new envoy.server.data.GroupMessage(groupMessage, Instant.now());
		// Telling the server to reload the message without the attachment and telling
		// the client not to send anymore attachments
		if (!config.isAttachmentSupportEnabled() && groupMessage.hasAttachment()) {
			groupMessage = groupMessageServer.toCommon();
			writeProxy.write(socketID, new NoAttachments());
		}

		// This is needed unfortunately because of f***ing lambda restrictions ("must be
		// fINaL oR EFfEcTivELy FiNAl")
		final var groupMessageCopy = groupMessage;
		// Deliver the message to the recipients that are online
		writeProxy.writeToOnlineContacts(
				persistenceManager.getGroupByID(groupMessageCopy.getRecipientID())
					.getContacts()
					.stream()
					.filter(c -> c.getID() != groupMessageCopy.getSenderID()),
				groupMessageCopy);

		try {
			PersistenceManager.getInstance().addMessage(groupMessageServer);
		} catch (final EntityExistsException e) {
			logger.warning("Received a groupMessage with an ID that already exists");
		}
	}
}
